import json

import openpyxl
from openpyxl.utils import get_column_letter
from openpyxl.styles import PatternFill, Font, Color, colors, Alignment, Border, Side

from simple_loggers import SimpleLogger

from pypubmed.util import safe_open


class Export(object):

    logger = SimpleLogger('Export')

    def __init__(self, data, outfile='out.xlsx', outtype=None, fields=None, **kwargs):
        self.outfile = outfile
        self.outtype = outtype or outfile.split('.')[-1]
        self.data = list(self.filter_data(data, fields)) if fields else data

    def filter_data(self, data, fields):
        fields = fields.strip(',').split(',')
        for context in data:
            out_ctx = {k: v for k, v in context.items() if k in fields}
            yield out_ctx

    def export(self):
        if self.outtype == 'xlsx':
            self.export_xlsx()
        elif self.outtype == 'json':
            self.export_json()
        elif self.outtype in ('jl', 'jsonlines'):
            self.export_json_lines()
        else:
            self.logger.error('outtype is invalid, please check!')
            exit(1)
        
        self.logger.info('save file: {}'.format(self.outfile))

    def write_title(self, sheet, titles, fg_color=colors.WHITE, bg_color=colors.BLACK, border=True, bold=True, width=18, size=12):
        for col, value in enumerate(titles, 1):

                w = width * 4 if value in ('abstract', 'abstract_cn') else width

                sheet.column_dimensions[get_column_letter(col)].width = w
                _ = sheet.cell(1, col, value=value)

                # set cell styles
                _.alignment = Alignment(horizontal='left',vertical='center',wrap_text=True)
                _.fill = PatternFill(start_color=bg_color, end_color=bg_color, fill_type="solid")
                _.font = Font(bold=bold, color=fg_color, size=size)

    def add_hyperlink(self, key, value):
        """
            method1:
                value=HYPERLINK("http://www.baidu.com", "baidu")
                sheet.cell(row, colum, value=value)

            method2:
                _ = sheet.cell(row, column, value='baidu')
                _.hyperlink = 'http://www.baidu.com'
        """
        if key == 'pmid':
            url = 'https://pubmed.ncbi.nlm.nih.gov/{value}/'
        elif key == 'pmc':
            url = 'https://www.ncbi.nlm.nih.gov/pmc/articles/{value}/'
        elif key == 'doi':
            url = 'https://doi.org/{value}'
        else:
            return None

        url = url.format(**locals())

        return url
    
    @property
    def all_fields(self):
        return '''
            pmid title abstract abstract_cn impact_factor journal med_abbr iso_abbr pubdate pubmed_pubdate
            pmc issn e_issn doi year pagination volume issue   
            pub_status authors keywords pub_types cited
        '''.split()

    def sort(self, item):
        if type(item) == tuple:
            k = item[0]
        else:
            k = item

        if k in self.all_fields:
            return self.all_fields.index(k)
        else:
            return 9999

    def export_xlsx(self, sheet_title='Result', freeze_panes='B1'):
        """
            PatternFill:
            - https://openpyxl.readthedocs.io/en/latest/api/openpyxl.styles.fills.html?highlight=PatternFill
        """
        book = openpyxl.Workbook()
        sheet = book.active

        # # freeze the first column
        # sheet.freeze_panes = 'B1'

        # # freeze the first row
        # sheet.freeze_panes = 'A2'

        # freeze the first column and the first row
        sheet.freeze_panes = 'B2'

        sheet.title = sheet_title

        titles = sorted(list(self.data[0].keys()), key=self.sort)
        self.write_title(sheet, titles)
        
        for row, context in enumerate(self.data, 2):

            color = '00b3ffb3' if row % 2 else '00b3ffff'
            for col, (key, value) in enumerate(sorted(list(context.items()), key=self.sort), 1):
                if type(value) == list:
                    try:
                        value = ', '.join(value)
                    except:
                        try:
                            value = json.dumps(value, ensure_ascii=False)
                        except:
                            value = str(value)
                elif type(value) == dict:
                    value = json.dumps(value, ensure_ascii=False)
                    
                _ = sheet.cell(row, col, value=value)
                _.fill = PatternFill(start_color=color, end_color=color, fill_type="solid")

                if key in ('pmid', 'pmc', 'doi') and value != '.':
                    _.hyperlink = self.add_hyperlink(key, value)
                    _.font = Font(color=colors.BLUE, italic=True)

                wrap_text = None
                if key in ('abstract', 'abstract_cn'):
                    wrap_text = True
                
                _.alignment = Alignment(horizontal='left',vertical='center',wrap_text=wrap_text)

        book.save(self.outfile)

    def export_json(self, **kwargs):
        with safe_open(self.outfile, 'w') as out:
            json.dump(self.data, out, ensure_ascii=False, **kwargs)

    def export_json_lines(self):
        with safe_open(self.outfile, 'w') as out:
            for context in self.data:
                out.write(json.dumps(context, ensure_ascii=False) + '\n')


if __name__ == '__main__':
    data = [
        {'pmid':1, 'pmc': 'PMC4472309', 'doi': '10.1016/0006-2952(75)90018-0'},
        {'pmid':1, 'pmc': '.', 'doi': '10.1016/0006-2952(75)90018-0'},
        {'pmid':1, 'pmc': 'PMC4472309', 'doi': '.'},
        {'pmid':1, 'pmc': 'PMC4472309', 'doi': '10.1016/0006-2952(75)90018-0'},
        {'pmid':1, 'pmc': 'PMC4472309', 'doi': '10.1016/0006-2952(75)90018-0'},
    ]

    Export(data).export()



import os
import json

import click

from pypubmed.core.eutils import Eutils
from pypubmed.core.citations import ncbi_citations, manual_citations


__epilog__ = '''
examples:\n
    citations 1 2 3\n
    citations 1 2 3 -f nlm\n
    citations 1 2 3 -m -f apa\n
'''

@click.command(name='citations',
               epilog=__epilog__,
               short_help='generate citations for given pmids',
               help='Citations Tools')
@click.option('-m', '--manual', help='cite with manual citations, default with ncbi citations', default=False, is_flag=True)
@click.option('-f', '--fmt', help='the format of citation', type=click.Choice('ama mla apa nlm'.split()), default='ama')
@click.option('-k', '--api-key', help='the API_KEY of NCIB', envvar='NCBI_API_KEY')
@click.argument('pmids', nargs=-1)
def citations_cli(pmids, manual, fmt, api_key):
    if not pmids:
        click.secho('please supply pmids or a file', fg='green')
        exit(1)

    if os.path.isfile(pmids[0]):
        pmid_list = open(pmids[0]).read().strip().split()
    else:
        pmid_list = pmids
    
    if manual:
        e = Eutils(api_key=api_key)
        articles = e.efetch(pmid_list)
        for article in articles:
            citation = manual_citations(article, fmt=fmt)
            click.secho(citation, fg='yellow')
    else:
        for pmid in pmid_list:
            citation = ncbi_citations(pmid, fmt=fmt)
            click.secho(json.dumps(citation, indent=2), fg='green')
    

if __name__ == '__main__':
    citations()



"""
    https://www.ncbi.nlm.nih.gov/books/NBK25497/
"""
import os
import sys
import json
import textwrap
import datetime

import click
import tqdm
import prettytable

from dateutil.parser import parse as date_parse

from impact_factor import ImpactFactor
from simple_googletrans import GoogleTrans
from simple_loggers import SimpleLogger
from webrequests import WebRequest


from pypubmed.util import xml_parser
from pypubmed.core.article import Article



class Eutils(object):
    """
        params:
            db          database name
            api_key     api_key or NCBI_API_KEY in environment

        optional params:
            term        term for esearch
            id          id(s) for efetch
            field       field for esearch
    """
    base_url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'

    logger = SimpleLogger('Eutils')

    def __init__(self, db='pubmed', api_key=None, **kwargs):
        self.db = db
        self.api_key = api_key or os.getenv('NCBI_API_KEY')
        self.validate_api_key()
        self.IF = ImpactFactor()
        self.TR = GoogleTrans()

    def parse_params(self, **kwargs):
        """
            - add default db
            - add api_key if available
        """
        params = {'db': self.db}
        if self.api_key:
            params['api_key'] = self.api_key
        params.update(kwargs)
        
        if 'api_key' in params and params['api_key'] is None:
            del params['api_key']

        return params

    def esearch(self, term, retstart=0, retmax=250, head=False, **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESearch

            > -search from a database from given term
            >> - esearch.cgi?db=pubmed&term=ngs
            >> - esearch.cgi?db=pubmed&term=ngs&retmode=xml&field=TIAB
            >> - esearch.cgi?db=pubmed&term=ngs[Title/Abstract]&retmode=xml
        """
        url = self.base_url + 'esearch.fcgi'
        params = self.parse_params(term=term, retmode='json', retstart=retstart, retmax=retmax, **kwargs)
        result = WebRequest.get_response(url, params=params).json()['esearchresult']

        if head:
            return result

        self.logger.info('{count} articles found with term: \x1b[32m{querytranslation}\x1b[0m'.format(**result))
        idlist = result['idlist']

        while int(result['retstart']) + int(result['retmax']) < int(result['count']):
            retstart = int(result['retstart']) + int(result['retmax'])
            params = self.parse_params(term=term, retmode='json', retstart=retstart, retmax=retmax, **kwargs)
            result = WebRequest.get_response(url, params=params).json()['esearchresult']
            idlist += result['idlist']

        if idlist:
            self.logger.info('idlist: {} ...'.format(', '.join(idlist[:10])))
        else:
            self.logger.warning('no result for term: {}'.format(term))
        return idlist

    def efetch(self, ids, batch_size=5, show_process=False, **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch

            > - fetch from a database for given ids
            >> efetch.cgi?db=pubmed&id=1,2,3
        """
        url = self.base_url + 'efetch.fcgi'

        self.logger.info('fetching start: total {}, batch_size: {}'.format(len(ids), batch_size))

        with tqdm.tqdm(total=len(ids), unit='articles', desc='Fetching', ncols=80, leave=False) as pbar:
        # with click.progressbar(length=len(ids), show_percent=True, empty_char=' ', fill_char='>') as pbar:
            for n in range(0, len(ids), batch_size):
                _id = ','.join(ids[n:n+batch_size])
                if show_process:
                    end =  n + batch_size if n + batch_size <= len(ids) else len(ids)
                    pbar.set_description('Fetching {}-{}'.format(n+1, end))
                    # pbar.label = 'Fetching {}-{}'.format(n+1, end)
                    pbar.update(batch_size)
                n += batch_size

                params = self.parse_params(id=_id, retmode='xml')

                xml = WebRequest.get_response(url, params=params).text
                
                for context in xml_parser.parse(xml):
                    article = Article(**context)
                    yield article

    def einfo(self, **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EInfo

            > - show all database list
            >> einfo.fcgi?db=
            > - show dbinfo for given database
            >> einfo.fcgi?db=pubmed
        """
        url = self.base_url + 'einfo.fcgi'
        params = self.parse_params(retmode='json', **kwargs)
        info = WebRequest.get_response(url, params=params).json()
        return info

    def elink(self, ids, dbfrom='pubmed', cmd='neighbor', **kwargs):
        """
            https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ELink

            > - get cited (`{"linkname": "pubmed_pubmed_citedin"}`)
            >> elink.fcgi?dbfrom=pubmed&db=pubmed&id=20210808&cmd=neighbor&retmode=json
            > - get pdf url
            >> elink.fcgi?dbfrom=pubmed&db=pubmed&cmd=prlinks&id=10210801

            cmds:
                - neighbor (default)
                - neighbor_score
                - neighbor_history
                - acheck
                - ncheck
                - lcheck
                - llinks
                - llinkslib
                - prlinks
        """
        url = self.base_url + 'elink.fcgi'
        params = self.parse_params(retmode='json', id=ids, dbfrom=dbfrom, cmd=cmd,  **kwargs)
        result = WebRequest.get_response(url, params=params).json()
        return result

    def get_cited(self, _id, dbfrom='pubmed', cmd='neighbor'):
        """
            get the cited pmids for given pmid
        """
        linksetdbs = self.elink(_id, dbfrom=dbfrom, cmd=cmd)['linksets'][0]['linksetdbs']
        links = [
            linkset['links'] for linkset in linksetdbs
            if linkset['linkname'] == 'pubmed_pubmed_citedin'
        ]
        
        citedin = links[0] if links else []
        return {'count': len(citedin), 'links': citedin}

    def get_pdf_url(self, _id, dbfrom='pubmed', cmd='prlinks'):
        """
            get the pdf url for given pmid
        """
        idurllist = self.elink(_id, dbfrom=dbfrom, cmd=cmd)['linksets'][0]['idurllist']
 
        result = {each['id']: each['objurls'][0]['url']['value'] for each in idurllist}

        return result

    def validate_fields(self, **kwargs):
        """
            Show all fields for given database
        """
        fieldlist = self.einfo(**kwargs)['einforesult']['dbinfo'][0]['fieldlist']

        fieldlist = sorted(fieldlist, key=lambda x: x.get('fullname'))

        fields = {}
        table = prettytable.PrettyTable(field_names=['Number', 'Name', 'FullName', 'Description'])
        for n, each in enumerate(fieldlist, 1):
            # print('{n}\t{name}\t{fullname}\t{description}'.format(n=n, **each))
            fields[n] = [each['name'], each['fullname'], each['description']]
            table.add_row([n, each['name'], each['fullname'], each['description']])

        table.align['FullName'] = 'l'
        table.align['Description'] = 'l'
        click.secho(str(table), fg='bright_blue')

        return fields

    def validate_api_key(self):
        """
            Description: https://www.ncbi.nlm.nih.gov/account/settings/#accountSettingsApiKeyManagement
                - E-utils users are allowed 3 requests/second without an API key.
                - Create an API key to increase your e-utils limit to 10 requests/second.
        """
        if not self.api_key:
            msg = textwrap.dedent('''
                API_KEY not found! Use eutils with a api_key is a good idea!
                You can create one from: https://www.ncbi.nlm.nih.gov/account/settings/#accountSettingsApiKeyManagement
                If you already have one, you can add it to your environment: export NCBI_API_KEY=xxxx''')
            self.logger.warning(msg)
            return

        res = self.einfo()
        if 'error' in res:
            self.logger.warning('invalid api_key, please check: {}'.format(self.api_key))
            self.api_key = None
        else:
            self.logger.info('\x1b[1;3;32mValid api_key: {}\x1b[0m'.format(self.api_key))

    def search(self, term, cited=True, translate=True, impact_factor=True, min_factor=None, **kwargs):
        """
            term:
                - string, eg. 'ngs AND disease'
                - pmid, eg. 1,2,3
                - file with pmid
        """
        if os.path.isfile(term):
            idlist = [open(term).read().split()]
        elif all(each.isdigit() for each in term.split(',')):
            idlist = term.split(',')
        else:
            idlist = self.esearch(term)

        # print(kwargs);exit()

        articles = self.efetch(idlist, **kwargs)
        for article in articles:
            if impact_factor:
                res = self.IF.search(article.issn) or self.IF.search(article.e_issn)
                article.impact_factor = res['factor'] if res else '.'

                # filter impact factor
                if min_factor and (article.impact_factor != '.' and article.impact_factor < min_factor):
                    continue

            if cited:
                article.cited = self.get_cited(article.pmid)
                
            if translate:
                article.abstract_cn = self.TR.translate(article.abstract)
            yield article



"""====================\n
NCBI Pubmed ToolKits\n
====================
"""
import os
import json

BASE_DIR = os.path.dirname(os.path.realpath(__file__))

version_info = json.load(open(os.path.join(BASE_DIR, 'version', 'version.json')))



if __name__ == '__main__':
    print(version_info)



import os
import sys
import logging

import click
import colorama

from ._search import search, advance_search
from ._citations import citations_cli
from pypubmed import version_info, __doc__


PY2 = sys.version_info.major == 2
if PY2:
    reload(sys)
    sys.setdefaultencoding('utf-8')


colorama.init()


log_level_maps = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}

__epilog__ = click.style('''
contact: {author} <{author_email}>
'''.format(**version_info), fg='bright_black')

@click.group(epilog=__epilog__, help=click.style(__doc__, fg='bright_blue', bold=True))
@click.option('-l', '--log-level',
              help='the mode of log',
              show_default=True,
              default='debug',
              type=click.Choice(log_level_maps.keys()))
@click.option('-k', '--api-key',
              help='the api_key of NCBI Pubmed, NCBI_API_KEY environment is available',
              envvar='NCBI_API_KEY',
              show_envvar=True)
@click.version_option(version=version_info['version'], prog_name='pypubmed')
@click.pass_context
def cli(ctx, **kwargs):
    kwargs['log_level'] = log_level_maps[kwargs['log_level']]
    ctx.obj = kwargs


def main():
    cli.add_command(search)
    cli.add_command(advance_search)
    cli.add_command(citations_cli)
    cli()


if __name__ == '__main__':
    main()



# -*- encoding: utf8 -*-
import os
import json
import codecs
from setuptools import setup, find_packages


BASE_DIR = os.path.dirname(os.path.abspath(__file__))
version_info = json.load(open(os.path.join(BASE_DIR, 'pypubmed', 'version', 'version.json')))


setup(
    name='pypubmed',
    version=version_info['version'],
    author=version_info['author'],
    author_email=version_info['author_email'],
    description=version_info['desc'],
    long_description=codecs.open(os.path.join(BASE_DIR, 'README.md'), encoding='utf-8').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/suqingdong/pypubmed',
    project_urls={
        'Documentation': 'https://pypubmed.readthedocs.io',
        'Tracker': 'https://github.com/suqingdong/pypubmed/issues',
    },
    license='BSD License',
    install_requires=codecs.open(os.path.join(BASE_DIR, 'requirements.txt'), encoding='utf-8').read().split('\n'),
    packages=find_packages(),
    include_package_data=True,
    entry_points={'console_scripts': [
        'pubmed = pypubmed.bin.cli:main',
        'pypubmed = pypubmed.bin.cli:main',
    ]},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries'
    ]
)
